const CONSTANTS = require('@src/utils/constants');
const displayUtil = require('@src/utils/display-utility');
const stringUtils = require('@src/utils/string-utils');
const { URL, URLSearchParams } = require('url');
const open = require('open');
const portscanner = require('portscanner');
const http = require('http');
const ui = require('./ui');

const SIGNIN_PATH = '/ap/signin';
const LOCALHOST_PORT = '9090';

module.exports = {
    solveCaptcha
};

/**
 * Navigate user to the captcha validation page
 *
 * @param needBrowser
 * @param captchaUrl
 * @param callback (error)
 * @return callback with error
 */
function solveCaptcha(needBrowser, captchaUrl, callback) {
    if (needBrowser) {
        console.log('Go to the CAPTCHA page, confirm that you are signed into the correct developer account, and solve the CAPTCHA.\n'
            + 'If your browser does not open the page, quit this process, paste the following url to your browser, '
            + `and complete the CAPTCHA there. Then try again.\n${captchaUrl}`);

        portscanner.checkPortStatus(LOCALHOST_PORT, (err, status) => {
            if (err) {
                callback(err);
            } else {
                if (status === 'closed') {
                    _openLoginUrlWithRedirectLink(captchaUrl);
                    _listenResponseFromCaptchaServer(LOCALHOST_PORT, (error) => {
                        callback(error);
                    });
                } else {
                    callback(`[Warn]: ${LOCALHOST_PORT} port on localhost has been occupied, `
                        + 'ask-cli cannot start a local server for receiving authorization code.\n'
                        + `Please either abort any processes running on port ${LOCALHOST_PORT} `
                        + 'or add `--no-browser` flag to the command as an alternative approach.');
                }
            }
        });
    } else {
        console.log(`Paste the following url to your browser and complete the CAPTCHA there. \n${captchaUrl}`);
        ui.getConfirmation((isCaptchaDone) => {
            if (isCaptchaDone) {
                callback();
            } else {
                console.log('You must complete the CAPTCHA validation before creating your first Alexa hosted skill.');
                process.exit(1);
                callback();
            }
        });
    }
};

/**
 * Build and open the url that navigates user to captcha validation page after login.
 * @param captcha validation url
 */
function _openLoginUrlWithRedirectLink(captchaUrl) {
    const envVarSignInHost = process.env.ASK_LWA_AUTHORIZE_HOST;
    const loginUrl = new URL(envVarSignInHost && stringUtils.isNonBlankString(envVarSignInHost)
        ? envVarSignInHost + SIGNIN_PATH : CONSTANTS.LWA.SIGNIN_URL);

    loginUrl.search = new URLSearchParams([
        ['openid.ns', 'http://specs.openid.net/auth/2.0'],
        ['openid.mode', 'checkid_setup'],
        ['openid.claimed_id', 'http://specs.openid.net/auth/2.0/identifier_select'],
        ['openid.identity', 'http://specs.openid.net/auth/2.0/identifier_select'],
        ['openid.assoc_handle', 'amzn_dante_us'],
        ['openid.return_to', `${captchaUrl}?redirect_url=http://127.0.0.1:9090/captcha`],
        ['openid.pape.max_auth_age', 7200]
    ]).toString();
    open(loginUrl.href);
}

/**
 * Start a local server and listen the response from the captcha validation server,
 * then extract validation result from it.
 * @param PORT
 * @param callback with error
 */
function _listenResponseFromCaptchaServer(PORT, callback) {
    let listenSpinner;
    const server = http.createServer(handleServerRequest);
    server.on('connection', (socket) => {
        socket.unref();
    });
    server.on('error', (error) => {
        callback(error);
    });
    server.listen(PORT, () => {
        listenSpinner = new displayUtil.Spinner(`Listening on http://localhost:${PORT}...`);
        listenSpinner.start();
    });

    function handleServerRequest(request, response) {
        response.on('close', () => {
            request.socket.destroy();
        });
        listenSpinner.stop();
        server.close();
        server.unref();
        if (request.url.startsWith('/captcha?success')) {
            response.end('CAPTCHA validation was successful. Please close the browser and return to the command line interface.');
            callback();
        } else if (request.url.startsWith('/captcha?error')) {
            const errorMsg = '[Error]: Failed to validate the CAPTCHA with internal service error. Please try again later.';
            response.statusCode = 500;
            response.end(errorMsg);
            callback(errorMsg);
        } else if (request.url.startsWith('/favicon.ico')) {
            request.destroy();
            response.statusCode = 204;
            response.end();
        } else {
            const errorMsg = '[Error]: Failed to validate the CAPTCHA. Please try again.';
            response.statusCode = 404;
            response.end(errorMsg);
            callback(errorMsg);
        }
    }
}
